@@ -1,10 +1,10 @@
 use strictures 2;
-package OpenAPI::Modern; # git description: v0.103-2-g9a0d4a31
+package OpenAPI::Modern; # git description: v0.104-11-ge2d9fc88
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Validate HTTP requests and responses against an OpenAPI v3.1 or v3.2 document
 # KEYWORDS: validation evaluation JSON Schema OpenAPI v3.1 v3.2 Swagger HTTP request response
 
-our $VERSION = '0.104';
+our $VERSION = '0.105';
 
 use 5.020;
 use utf8;
@@ -35,7 +35,7 @@ use OpenAPI::Modern::Utilities qw(add_vocab_and_default_schemas);
 use JSON::Schema::Modern::Document::OpenAPI;
 use MooX::TypeTiny 0.002002;
 use Types::Standard qw(InstanceOf Bool);
-use Mojo::Util 'url_unescape';
+use Mojo::Util qw(url_unescape punycode_decode);
 use Mojo::Message::Request;
 use Mojo::Message::Response;
 use Storable 'dclone';
@@ -439,10 +439,7 @@ sub find_path_item ($self, $options, $state = {}) {
   if (not $options->{path_template} and $options->{uri}) {
     # derive path_template and capture values from the request URI
 
-    # sorting (ascii-wise) gives us the desired results that concrete path components sort ahead of
-    # templated components, except when the concrete component is a non-ascii character or matches
-    # 0x7c (pipe), 0x7d (close-brace) or 0x7e (tilde)
-    foreach my $pt (sort keys(($schema->{paths}//{})->%*)) {
+    foreach my $pt ($self->openapi_document->path_templates->@*) {
       my $local_state = +{ %$state };
       $local_state->{path_item} = $schema->{paths}{$pt};
       $local_state->{keyword_path} = jsonp('/paths', $pt);
@@ -546,9 +543,7 @@ sub find_path_item ($self, $options, $state = {}) {
   $options->{operation_id} = $options->{_operation}{operationId}
     if exists $options->{_operation}{operationId};
 
-  # note: we aren't doing anything special with escaped slashes. this bit of the spec is hazy.
-  # { for the editor
-  my @path_capture_names = ($options->{path_template} =~ m!\{([^}]+)\}!g);
+  my @path_capture_names = ($options->{path_template} =~ /\{([^{}]+)\}/g);
   return E({ %$state, $options->{uri} ? (data_path => '/request/uri') : (), recommended_response => [ 500 ] }, 'provided path_captures names do not match path template "%s"', $options->{path_template})
     if exists $options->{path_captures}
       and not is_equal([ sort keys $options->{path_captures}->%* ], [ sort @path_capture_names ]);
@@ -619,24 +614,31 @@ sub recursive_get ($self, $uri_reference, $entity_type = undef) {
 # consideration additional information in the current path-item), and extract capture values.
 # returns false on error, possibly adding errors to $state.
 sub _match_uri ($self, $method, $uri, $path_template, $state) {
-  my $uri_path = $uri->path->to_string;
+  $uri = $uri->clone->fragment(undef)->query('');
 
   # RFC9112 §3.2.1-3: "If the target URI's path component is empty, the client MUST send "/" as the
-  # path within the origin-form of request-target."
-  $uri_path = '/' if not length $uri_path;
+  # path within the origin-form of request-target." This also lets us match a path template of "/".
+  $uri->path->leading_slash(1); # no effect on stringified URI unless path is empty
 
   # v3.2.0 §4.8.2, "Path Templating": "The value for these path parameters MUST NOT contain any
   # unescaped “generic syntax” characters described by [RFC3986]: forward slashes (/), question
   # marks (?), or hashes (#)."
   my $path_pattern = join '',
     map +(substr($_, 0, 1) eq '{' ? '([^/?#]*)' : quotemeta($_)), # { for the editor
-    split /(\{[^}]+\})/, $path_template;
+    split /(\{[^{}]+\})/, $path_template;
 
   # if the uri doesn't match against the path alone, we can immediately bail (and keep looking for
   # another /paths entry that might match)... this also saves us needless parsing of server objects
   do { use autovivification 'store'; push $state->{debug}{uri_patterns}->@*, $path_pattern.'$' }
     if exists $state->{debug};
-  return if $uri_path !~ m/$path_pattern$/;
+  return if $uri !~ m/$path_pattern$/;
+
+  # identify the unmatched part of the request URI, to be later matched against server urls
+  my $uri_prefix = substr($uri, 0, -length($&));
+
+  # extract all capture values from path template variables: ($1 .. $n)
+  # perldoc perlvar, @-: $n coincides with "substr $_, $-[n], $+[n] - $-[n]" if "$-[n]" is defined
+  my @path_capture_values = map _uri_decode(substr($uri, $-[$_], $+[$_]-$-[$_])), 1 .. $#-;
 
   # we set aside $state for potential restoration because we might still encounter issues later on
   # that require us to keep iterating for another URI match
@@ -660,8 +662,6 @@ sub _match_uri ($self, $method, $uri, $path_template, $state) {
 
   return if not $local_state->{operation};
 
-  my $full_uri = $uri->clone->path($uri_path)->fragment(undef)->query('')->to_string;
-
   # we need to keep track of the traversed path to the servers object, as well as its absolute
   # location, for usage in error objects
   my ($servers, $more_keyword_path, $base_schema_uri) =
@@ -678,41 +678,48 @@ sub _match_uri ($self, $method, $uri, $path_template, $state) {
   # url value of `/`."
   $servers = [{ url => '/' }] if not $servers or not @$servers;
 
+  my @path_capture_names = ($path_template =~ /\{([^{}]+)\}/g);
+
   foreach my $index_and_server (pairs indexed @$servers) {
     my ($index, $server) = @$index_and_server;
 
-    # we need a full uri to match against the full uri taken from the request (scheme and host)
+    # We need a full uri to match against the full uri taken from the request (scheme and host)
     # we fall back to using the request's scheme, host and port, otherwise the match can never
     # succeed.
-    my $server_url = Mojo::URL->new($server->{url})->to_abs($self->openapi_document->retrieval_uri);
-    $server_url = $server_url->to_abs($uri)->query('');
-    $server_url->path->leading_slash(0)->trailing_slash(0);
+    # But before we apply URI logic to the server url, we must first protect any templated sections
+    # so they are not altered by normalization. We use NUL, as it is unchanged in the host during
+    # punycode-encoding
+    my $normalized_server_url = Mojo::URL->new($server->{url} =~ s/\{[^{}]+\}/\x00/gr)
+      ->to_abs($self->openapi_document->retrieval_uri)
+      ->to_abs($uri);
 
-    # turn %7B and %7D back into { and }
-    my $uri_template = url_unescape($server_url) . $path_template;
+    # strips slash if path is '/'; otherwise has no effect on stringified URI
+    $normalized_server_url->path->leading_slash(0);
 
-    my $uri_pattern = join '',
-      map +(substr($_, 0, 1) eq '{' ? '([^/?#]*)' : quotemeta($_)), # { for the editor
-      split /(\{[^}]+\})/, $uri_template;
-    do { use autovivification 'store'; push $state->{debug}{uri_patterns}->@*, '^'.$uri_pattern.'$' }
+    my $server_pattern = join '',
+      map +($_ eq '%00' ? '([^/?#]*)' : quotemeta($_)),
+      split /(%00)/, $normalized_server_url;  # all NULs appear as %00 in the stringified form
+    do { use autovivification 'store'; push $state->{debug}{uri_patterns}->@*, '^'.$server_pattern }
       if exists $state->{debug};
-    next if $full_uri !~ m/^$uri_pattern$/;
+    next if $uri_prefix !~ m/^$server_pattern$/;
 
-    # extract all capture values from server variables and path template variables: ($1 .. $n)
+    # extract all capture values from server variables: ($1 .. $n)...
     # perldoc perlvar, @-: $n coincides with "substr $_, $-[n], $+[n] - $-[n]" if "$-[n]" is defined
-    my @uri_capture_values = map
-      Encode::decode('UTF-8', url_unescape(substr($full_uri, $-[$_], $+[$_]-$-[$_])),
-        Encode::DIE_ON_ERR), 1 .. $#-;
+    my @server_capture_values = map substr($uri_prefix, $-[$_], $+[$_]-$-[$_]), 1 .. $#-;
+
+    # ...and punycode-decode those from the host, and url-unescape those from the path
+    my $host_variable_count = ()= ($normalized_server_url->host//'') =~ /\x00/g;
+    @server_capture_values = (
+      (map +(/^xn--(.+)$/ ? punycode_decode($1) : $_), @server_capture_values[0 .. $host_variable_count-1]),
+      (map _uri_decode($_), @server_capture_values[$host_variable_count .. $#server_capture_values]));
 
     # we have a match, so preserve our new $state values created via _resolve_ref
     %$state = %$local_state;
 
-    # note: we aren't doing anything special with escaped slashes. this bit of the spec is hazy.
-    # { for the editor
-    my @uri_capture_names = ($uri_template =~ m!\{([^}]+)\}!g);
+    my @server_capture_names = ($server->{url} =~ /\{([^{}]+)\}/g);
 
     my ($valid, %seen) = (1);
-    foreach my $name (@uri_capture_names) {
+    foreach my $name (@server_capture_names, @path_capture_names) {
       # TODO: ideally this should be caught at document load time, but the use of $refs between
       # /paths entries and path-items makes this difficult
       $valid = E({ %$state, keyword => 'url', data_path => '/request/uri',
@@ -725,9 +732,9 @@ sub _match_uri ($self, $method, $uri, $path_template, $state) {
     return if not $valid;
 
     my %captures;
-    @captures{@uri_capture_names} = @uri_capture_values;
+    @captures{@server_capture_names} = @server_capture_values;
 
-    foreach my $name (@uri_capture_names) {
+    foreach my $name (@server_capture_names) {
       next if not exists((($server->{variables}//{})->{$name}//{})->{enum});
 
       $valid = E({ %$state, data_path => '/request/uri', keyword => 'enum',
@@ -739,6 +746,8 @@ sub _match_uri ($self, $method, $uri, $path_template, $state) {
         if not any { $captures{$name} eq $_ } $server->{variables}{$name}{enum}->@*;
     }
 
+    @captures{@path_capture_names} = @path_capture_values;
+
     return if not $valid;
     return \%captures;
   }
@@ -905,6 +914,7 @@ sub _validate_querystring_parameter ($self, $state, $param_obj, $uri) {
     if $param_obj->{required} and not length $uri->query->{string};
 
   # Replace "+" with whitespace, unescape and decode as in Mojo::Parameters::pairs
+  # We do not UTF-8-decode the content: this is the responsibility of the media-type decoder.
   my $content = url_unescape($uri->query->{string} =~ s/\+/ /gr);
 
   $self->_validate_parameter_content({ %$state, depth => $state->{depth}+1 }, $param_obj, \$content)
@@ -913,7 +923,7 @@ sub _validate_querystring_parameter ($self, $state, $param_obj, $uri) {
 sub _validate_parameter_content ($self, $state, $param_obj, $content_ref) {
   my ($media_type) = keys $param_obj->{content}->%*;  # there can only be one key
 
-  # FIXME: handle media-type parameters better when selecting for a decoder, see RFC9110 8.3.1
+  # FIXME: handle media-type parameters better when selecting for a decoder, see RFC9110 §8.3.1
 
   if ($media_type =~ m{^text/}
       and my $charset = ($media_type =~ /\bcharset\s*=\s*"?([^"\s;]+)"?/i ? $1 : undef)) {
@@ -940,7 +950,7 @@ sub _validate_body_content ($self, $state, $content_obj, $message) {
       'missing header: Content-Type')
     if not length $content_type;
 
-  # FIXME: needs to handle media-type parameters when selecting for a decoder, see RFC9110 8.3.1
+  # FIXME: needs to handle media-type parameters when selecting for a decoder, see RFC9110 §8.3.1
 
   my $media_type = (first { fc($content_type) eq fc } keys $content_obj->%*)
     // (first { m{([^/]+)/\*$} && fc($content_type) =~ m{^\F\Q$1\E/[^/]+$} } keys $content_obj->%*);
@@ -1188,6 +1198,11 @@ sub _convert_response ($response) {
   return $res;
 }
 
+# url-percent-decode and UTF-8-decode a string
+sub _uri_decode ($str) {
+  Encode::decode('UTF-8', url_unescape($str), Encode::DIE_ON_ERR);
+}
+
 # callback hook for Sereal::Encoder
 sub FREEZE ($self, $serializer) { +{ %$self } }
 
@@ -1217,7 +1232,7 @@ OpenAPI::Modern - Validate HTTP requests and responses against an OpenAPI v3.1 o
 
 =head1 VERSION
 
-version 0.104
+version 0.105
 
 =head1 SYNOPSIS
 
@@ -1499,7 +1514,7 @@ C<operation_id>: a string corresponding to the L<operationId|https://learn.opena
 
 =item *
 
-C<path_captures>: a hashref mapping placeholders in the path template to their actual values in the request URI
+C<path_captures>: a hashref mapping placeholders in the path template to their actual values in the request URI; these will be url-unescaped
 
 =back
 
@@ -1515,11 +1530,13 @@ a L<JSON::Schema::Modern::Error> object, and the return value is false.
 
 In addition, these values are also populated in the options hash (when available):
 
+=for stopwords punycode
+
 =over 4
 
 =item *
 
-C<uri_captures>: a hashref mapping placeholders in the entire uri template (server url plus path template) to their actual values in the request URI
+C<uri_captures>: a hashref mapping placeholders in the entire uri template (server url plus path template) to their actual values in the request URI; these will be url-unescaped (for values coming from the path) and punycode-decoded (for values coming from the host).
 
 =item *
 
